/*  Screem:  screem-session.c
 *
 *  The ScreemSession object, inspired by session.c from Epiphany
 *
 *  Copyright (C) 2003 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include "config.h"

#include "screem-session.h"

#include "screem-page.h"
#include "pageUI.h"
#include "screem-site.h"
#include "screem-site-ui.h"
#include "screem-window.h"

#include "fileops.h"

#include <string.h>
#include <libgnomeui/gnome-client.h>

#include <libxml/tree.h>

static gchar *get_session_filename( const gchar *filename );
static gint
save_yourself( GnomeClient *client, gint phase, GnomeSaveStyle save_style,
               gint is_shutdown, GnomeInteractStyle interact_style,
               gint is_fast, ScreemSession *session );

enum {
	PROP_0,
	PROP_APP
};

struct ScreemSessionPrivate {
	ScreemApplication *app;
	GnomeClient *client;

	GString *output;

	/* hash table of classes and pos / dimensions 
	 * (name / GdkRectangle)*/
	GHashTable *dialogs;

	/* site file browser expansion status 
	 * (name / GHashTable (path/1) ) */
	GHashTable *site_expansion;
};

G_DEFINE_TYPE( ScreemSession, screem_session, G_TYPE_OBJECT )

static void screem_session_finalize( GObject *object );
static void screem_session_set_prop( GObject *object, guint prop_id,
				const GValue *value, GParamSpec *spec );
static void screem_session_get_prop( GObject *object, guint prop_id,
				GValue *value, GParamSpec *spec );

static void screem_session_class_init( ScreemSessionClass *klass )
{
	GObjectClass *obj_class;
	GParamSpec *spec;

	obj_class = G_OBJECT_CLASS( klass );
	obj_class->finalize = screem_session_finalize;
	obj_class->get_property = screem_session_get_prop;
	obj_class->set_property = screem_session_set_prop;

	spec = g_param_spec_object( "app", "Application",
				    "The Application",
				    G_TYPE_OBJECT,
				    G_PARAM_READWRITE |
				    G_PARAM_CONSTRUCT );
	g_object_class_install_property( obj_class, PROP_APP, spec );
}
	
static void screem_session_init( ScreemSession *session )
{
	ScreemSessionPrivate *priv;
	GnomeClient *client;
	
	priv = session->priv = g_new0( ScreemSessionPrivate, 1 );

	client = priv->client = gnome_master_client ();
        g_signal_connect( G_OBJECT( client ), "save_yourself",
			  G_CALLBACK( save_yourself ), session );
        g_signal_connect( G_OBJECT( client ), "die",
			  G_CALLBACK( gtk_main_quit ), NULL );

	priv->dialogs = g_hash_table_new_full( g_str_hash, g_str_equal,
			(GDestroyNotify)g_free,
			(GDestroyNotify)g_free );

	priv->site_expansion = g_hash_table_new_full( g_str_hash, g_str_equal,
			(GDestroyNotify)g_free,
			(GDestroyNotify)g_hash_table_destroy );
}

static void screem_session_finalize( GObject *object )
{
	ScreemSession *session;
	ScreemSessionPrivate *priv;
			
	g_return_if_fail( object != NULL );
	g_return_if_fail( SCREEM_IS_SESSION( object ) );

	session = SCREEM_SESSION( object );

	priv = session->priv;
	
	g_hash_table_destroy( priv->dialogs );
	g_hash_table_destroy( priv->site_expansion );

	g_free( priv );
	
	G_OBJECT_CLASS( screem_session_parent_class )->finalize( object );
}

static void screem_session_set_prop( GObject *object, guint prop_id,
				const GValue *value, GParamSpec *spec )
{
	ScreemSession *session;
	ScreemSessionPrivate *priv;
	
	session = SCREEM_SESSION( object );
	priv = session->priv;

	switch( prop_id ) {
		case PROP_APP:
			priv->app = g_value_get_object( value );
			break;
		default:
			break;
	}
}

static void screem_session_get_prop( GObject *object, guint prop_id,
				GValue *value, GParamSpec *spec )
{
	ScreemSession *session;
	ScreemSessionPrivate *priv;
	
	session = SCREEM_SESSION( object );
	priv = session->priv;

	switch( prop_id ) {
		case PROP_APP:
			g_value_set_object( value, priv->app );
			break;
		default:
			break;
	}
}

/* static stuff */
static gchar *get_session_filename( const gchar *filename )
{
	gchar *ret;

	g_return_val_if_fail( filename != NULL, NULL );

	if( ! strcmp( filename, SESSION_LAST ) ) {
		gchar *tmp;

		tmp = screem_get_dot_dir();
		ret = g_build_filename( tmp, "session_last.xml", NULL );
		g_free( tmp );
	} else if( ! strcmp( filename, SESSION_GNOME ) ) {
		gchar *tmp;

		tmp = screem_get_dot_dir();
		ret = g_build_filename( tmp, "session_gnome-XXXXXX", NULL );
		g_free( tmp );
		
		tmp = screem_create_tmp_file( ret, "xml", "" );
		g_free( ret );
		ret = tmp;
	} else {
		ret = g_strdup( filename );
	}
	
	return ret;
}

static gint
save_yourself( GnomeClient *client, gint phase, GnomeSaveStyle save_style,
               gint is_shutdown, GnomeInteractStyle interact_style,
               gint is_fast, ScreemSession *session )
{
	gchar *argv[] = { "screem", "--load-session", NULL };
        gchar *discard_argv[] = { "rm", "-f", NULL, NULL };

	argv[ 2 ] = get_session_filename( SESSION_GNOME );
	gnome_client_set_restart_command( client, 3, argv );
	
	discard_argv[ 2 ] = argv[ 2 ];
        gnome_client_set_discard_command( client, 3, discard_argv );
	
	if( ! screem_session_save( session, argv[ 2 ] ) ) {
		g_warning( "Failed to save session\n" );
	}
	
        return TRUE;
}

static gboolean save_expansion( gpointer key, gpointer value,
		gpointer data )
{
	GString *output;

	output = (GString*)data;
	
	g_string_append_printf( output, "<uri>%s</uri>",
			(const gchar*)key );
	
	return FALSE;
}

static void screem_session_save_site( ScreemSite *site, ScreemSession *session )
{
	ScreemApplication *app;
	GList *windows;
	GList *tmp;
	GString *output;
	const gchar *pathname;
	gchar *etmp;

	GHashTable *expansion;
	
	g_return_if_fail( SCREEM_IS_SESSION( session ) );
	g_return_if_fail( SCREEM_IS_SITE( site ) );
	g_return_if_fail( session->priv->output != NULL );
	
	g_object_get( G_OBJECT( session ), "app", &app, NULL );
	
	output = session->priv->output;
	
	pathname = screem_site_get_pathname( site );
	if( ! pathname ) {
		pathname = "";
	}
	
	etmp = g_markup_escape_text( pathname, -1 );
	g_string_append_printf( output, 
				"<site url=\"%s\">", etmp );
	g_free( etmp );
	
	windows = screem_application_get_window_list( app );
	for( tmp = windows; tmp; tmp = tmp->next ) {
		ScreemWindow *window;
		const gchar *name;
		GList *pages;
		GList *ptmp;

		window = SCREEM_WINDOW( tmp->data );
		name = "fixme";
		pages = screem_site_get_documents( site, name );
		etmp = g_markup_escape_text( name, -1 );
		g_string_append_printf( output, "<pages name=\"%s\">",
					etmp );
		g_free( etmp );
		
		for( ptmp = pages; ptmp; ptmp = ptmp->next ) {
			ScreemPage *page;

			page = SCREEM_PAGE( ptmp->data );
			pathname = screem_page_get_pathname( page );
			if( ! pathname ) {
				pathname = "";
			}
			etmp = g_markup_escape_text( pathname, -1 );
			g_string_append_printf( output, "<page url=\"%s\"/>",
						etmp );
			g_free( etmp );
		}
		g_string_append_printf( output, "</pages>" );
	}

	expansion = g_hash_table_lookup( session->priv->site_expansion,
			screem_site_get_name( site ) );
	if( expansion && g_hash_table_size( expansion ) ) {
		g_string_append( output, "<expanded>" );
		g_hash_table_foreach( expansion,
				(GHFunc)save_expansion,
				output );
		g_string_append( output, "</expanded>" );
	}
	
	g_string_append( output, "</site>" );
	g_object_unref( app );
}

static void screem_session_load_site( ScreemSession *session, 
					ScreemSite *site,
					xmlNodePtr node )
{
	ScreemApplication *app;
	gboolean fake;
	xmlNodePtr child;
	xmlChar *url;
	const gchar *sname;
	
	g_object_get( G_OBJECT( session ), "app", &app, NULL );
	
	fake = screem_site_get_fake_flag( site );
	sname = screem_site_get_name( site );
	
	while( node ) {
		if( ! strcmp( "expanded", (gchar*)node->name ) ) {
			/* nodes in the file browser to be
			 * expanded for this site */
			for( child = node->children; child;
					child = child->next ) {
				if( ! strcmp( "uri", (gchar*)child->name ) ) {
					url = xmlNodeGetContent( child );
					screem_session_site_set_expanded( session, sname, (const gchar *)url, TRUE );
					xmlFree( url );
				} 
			}
		} else if( ! strcmp( "pages", (gchar*)node->name ) ) {
			/* pages to be opened in window name="" */
			xmlChar *name;

			name = xmlGetProp( node, (const xmlChar*)"name" );
			if( name && *name != '\0' ) {
				GList *docs;
				ScreemPage *page;
				
				child = node->children;
				for( docs = NULL; child; child = child->next ) {
					if( ! strcmp( "page", (gchar*)child->name ) ) {
						url = xmlGetProp( child, (const xmlChar*)"url" );
						if( url ) {
							page = screem_site_locate_page( site, (gchar*)url );
							if( *url == '\0' && ! page ) {
								page = screem_page_new( G_OBJECT( app ) );
								screem_site_add_page( site, page );
								g_object_unref( page );
							} else if( fake && ! page ) {
								page = screem_page_new( G_OBJECT( app ) );
								screem_page_set_pathname( page, (gchar*)url );
								screem_site_add_page( site, page );
								g_object_unref( page );
							}
							xmlFree( url );
						} else {
							page = NULL;
						}
						if( page ) {
							docs = g_list_prepend( docs, page );
						}
					}
				}
				docs = g_list_reverse( docs );
				/* FIXME: name gets leaked */
				screem_site_set_documents( site, 
						(gchar*)name, docs );
			} else if( name ) {
				xmlFree( name );
			}
		}
		node = node->next;
	}
	g_object_unref( app );
}

static void screem_session_save_windows( ScreemWindow *window, 
					 ScreemSession *session )
{
	GString *output;
	const gchar *name;
	gint x;
	gint y;
	gint w;
	gint h;
	gboolean max;

	ScreemApplication *app;
	ScreemSite *site;
	const gchar *sname;
	gchar *tmp;
	
	g_return_if_fail( SCREEM_IS_SESSION( session ) );
	g_return_if_fail( SCREEM_IS_WINDOW( window ) );
	g_return_if_fail( session->priv->output != NULL );
	
	output = session->priv->output;
	
	g_object_get( G_OBJECT( session ), "app", &app, NULL );

	name = "fixme";
	
	gtk_window_get_position( GTK_WINDOW( window ), &x, &y );
	gtk_window_get_size( GTK_WINDOW( window ), &w, &h );
	max = window->max;
	if( max ) {
		x = window->x;
		y = window->y;
		w = window->w;
		h = window->h;
	} 
	g_string_append_printf( output, 
				"<window name=\"%s\" x=\"%i\" y=\"%i\" w=\"%i\" h=\"%i\" max=\"%i\">",
				name, x, y, w, h, max );
	
	
	site = screem_window_get_current( window );
	sname = screem_site_get_name( site );

	if( site == screem_application_get_default_site( app ) ) {
		sname = "";
	}
	tmp = g_markup_escape_text( sname, -1 );
	g_string_append_printf( output, "<site name=\"%s\" />",
				tmp );
	g_free( tmp );
	
	g_string_append( output, "</window>" );

	g_object_unref( app );
}

static void screem_session_load_window( ScreemSession *session,
					ScreemWindow *window,
					xmlNodePtr node )
{
	ScreemApplication *app;
	ScreemSite *dsite;
	ScreemSite *site;
	
	g_object_get( G_OBJECT( session ), "app", &app, NULL );
	site = NULL;
				
	dsite = screem_application_get_default_site( app );

	while( node ) {
		if( ! strcmp( "site", (gchar*)node->name ) ) {
			xmlChar *tmp;

			tmp = xmlGetProp( node, (const xmlChar*)"name" );
			if( ( ! tmp ) || *tmp == '\0' ) {
				site = dsite;
			} else {
				site = screem_application_get_site_by_name( app, (gchar*)tmp );
			}
			if( tmp ) {
				xmlFree( tmp );
			}
			if( site ) {
				screem_window_set_current( window, site );
				node = node->children;
			} else {
				node = node->next;
			}
		} else {
			node = node->next;
		}
	}
	/* if we failed to load any of the sites, set
	 * the default site as the current one */
	if( ! screem_window_get_current( window ) ) {
		screem_window_set_current( window, dsite );
	}
	g_object_unref( app );
}

static void screem_session_save_hint( ScreemSession *session )
{
	ScreemApplication *app;
	GtkWidget *hint;
	GString *output;
	gint x;
	gint y;
	
	g_return_if_fail( SCREEM_IS_SESSION( session ) );
	g_return_if_fail( session->priv->output != NULL );
	
	output = session->priv->output;
	
	g_object_get( G_OBJECT( session ), "app", &app, NULL );

	hint = app->hint;
	
	gtk_window_get_position( GTK_WINDOW( hint ), &x, &y );

	g_string_append_printf( output, "<hints x=\"%i\" y=\"%i\"/>",
				x, y );
	
	g_object_unref( app );
}

static void screem_session_load_hint( ScreemSession *session, 
		xmlNodePtr node )
{
	ScreemApplication *app;
	GtkWidget *hint;
	xmlChar *tmp;
	gint x;
	gint y;
				
	tmp = xmlGetProp( node, (const xmlChar*)"x" );
	x = atoi( (gchar*)tmp );
	xmlFree( tmp );
	tmp = xmlGetProp( node, (const xmlChar*)"y" );
	y = atoi( (gchar*)tmp );
	xmlFree( tmp );

	g_object_get( G_OBJECT( session ), "app", &app, NULL );

	hint = app->hint;
	gtk_window_set_position( GTK_WINDOW( hint ), 
				 GTK_WIN_POS_NONE );
	gtk_window_move( GTK_WINDOW( hint ), x, y );

	g_object_unref( app );
}

static gboolean save_dialog( gpointer key, gpointer value,
		gpointer data )
{
	GString *output;
	const gchar *name;
	GdkRectangle *rect;

	name = (const gchar *)key;
	rect = (GdkRectangle*)value;
	output = (GString*)data;

	g_string_append_printf( output, "<dialog name=\"%s\" x=\"%i\" y=\"%i\" w=\"%i\" h=\"%i\" />", name, rect->x, rect->y, rect->width, rect->height );

	return FALSE;
}

static void screem_session_save_dialogs( ScreemSession *session )
{
	ScreemSessionPrivate *priv;
	
	g_return_if_fail( SCREEM_IS_SESSION( session ) );
	g_return_if_fail( session->priv->output != NULL );
	
	priv = session->priv;
	g_hash_table_foreach( priv->dialogs, (GHFunc)save_dialog,
			priv->output );
}

static void screem_session_load_dialog( ScreemSession *session, 
		xmlNodePtr node )
{
	ScreemSessionPrivate *priv;
	xmlChar *tmp;
	gchar *name;
	GdkRectangle *rect;
	gint x;
	gint y;
	gint w;
	gint h;
	
	priv = session->priv;
	
	name = (gchar*)xmlGetProp( node, (const xmlChar*)"name" );
	tmp = xmlGetProp( node, (const xmlChar*)"x" );
	x = atoi( (gchar*)tmp );
	xmlFree( tmp );
	tmp = xmlGetProp( node, (const xmlChar*)"y" );
	y = atoi( (gchar*)tmp );
	xmlFree( tmp );
	tmp = xmlGetProp( node, (const xmlChar*)"w" );
	w = atoi( (gchar*)tmp );
	xmlFree( tmp );
	tmp = xmlGetProp( node, (const xmlChar*)"h" );
	h = atoi( (gchar*)tmp );
	xmlFree( tmp );

	rect = g_new0( GdkRectangle, 1 );
	rect->x = x;
	rect->y = y;
	rect->width = w;
	rect->height = h;
	g_hash_table_insert( priv->dialogs, name, rect );
}

/* public stuff */

ScreemSession *screem_session_new( GObject *app )
{
	ScreemSession *session;

	session = g_object_new( SCREEM_TYPE_SESSION, "app", app, NULL );

	return session;
}

void screem_session_load( ScreemSession *session,
			  const gchar *filename )
{
	ScreemApplication *app;
	xmlDocPtr doc;
	gchar *file;

	g_object_get( G_OBJECT( session ), "app", &app, NULL );
	
	file = get_session_filename( filename );
	doc = NULL;
	if( uri_exists( file, NULL ) ) {
		doc = xmlParseFile( file );
	}
	g_free( file );

	if( doc ) {
		xmlNodePtr node;

		node = xmlDocGetRootElement( doc );
		/* make sure we are dealing with a screem session file */
		if( strcmp( "screemsession", (gchar*)node->name ) ) {
			node = NULL;
		} else {
			node = node->children;
		}
		
		while( node ) {
			if( ! strcmp( "site", (gchar*)node->name ) ) {
				ScreemSite *site;
				xmlChar *tmp;

				tmp = xmlGetProp( node, (const xmlChar*)"url" );
				if( ( ! tmp ) || *tmp == '\0' ) {
					site = screem_application_get_default_site( app );
					g_assert( site );
				} else if( screem_site_open_with_filename( NULL, app, (gchar*)tmp ) ) {
					site = screem_application_get_site( app, (gchar*)tmp );
				} else {
					site = NULL;
				}
				if( tmp ) {
					xmlFree( tmp );
				}
				if( site ) {
					screem_session_load_site( session, site,
								  node->children );
				}
			} else if( ! strcmp( "window", (gchar*)node->name ) ) {
				ScreemWindow *win;
				xmlChar *tmp;
				gint x;
				gint y;
				gint w;
				gint h;
				gint max;
				
				tmp = xmlGetProp( node, (const xmlChar*)"x" );
				x = atoi( (gchar*)tmp );
				xmlFree( tmp );
				tmp = xmlGetProp( node, (const xmlChar*)"y" );
				y = atoi( (gchar*)tmp );
				xmlFree( tmp );
				tmp = xmlGetProp( node, (const xmlChar*)"w" );
				w = atoi( (gchar*)tmp );
				xmlFree( tmp );
				tmp = xmlGetProp( node, (const xmlChar*)"h" );
				h = atoi( (gchar*)tmp );
				xmlFree( tmp );
				tmp = xmlGetProp( node, (const xmlChar*)"max" );
				max = 0;
				if( tmp ) {
					max = atoi( (gchar*)tmp );
					xmlFree( tmp );
				}

				win = screem_application_create_window( app );
				win->x = x;
				win->y = y;
				win->w = w;
				win->h = h;
				win->max = max;

				screem_session_load_window( session, win, 
							node->children );
				
				gtk_window_set_default_size( GTK_WINDOW( win ),
								w, h );
				gtk_widget_show( GTK_WIDGET( win ) );
				gtk_window_move( GTK_WINDOW( win ), x, y );
				gtk_window_resize( GTK_WINDOW( win ),
						w, h );
				if( max ) {
					gtk_window_maximize( GTK_WINDOW( win ) );
				}
			} else if( ! strcmp( "hints", (gchar*)node->name ) ) {
				screem_session_load_hint( session, node );
			} else if( ! strcmp( "dialog", (gchar*)node->name ) ) {
				screem_session_load_dialog( session,
						node );
			}

			node = node->next;
		}
		
		xmlFreeDoc( doc );
	}
	g_object_unref( app );
}

gboolean screem_session_save( ScreemSession *session, const gchar *filename )
{
	ScreemSessionPrivate *priv;
	gboolean ret;
	GList *list;
	gchar *file;	
	GError *error;
	
	g_return_val_if_fail( SCREEM_IS_SESSION( session ), FALSE );

	priv = session->priv;

	/* ensure the the current site has a record of which pages
	 * are open in each window */
	list = screem_application_get_window_list( priv->app );
	while( list ) {
		ScreemSite *site;
		ScreemWindow *window;
		GList *docs;
		const gchar *name;
		
		window = SCREEM_WINDOW( list->data );
		name = "fixme";

		site = screem_window_get_current( window );
		g_assert( site );
		docs = screem_window_get_documents( window );
		screem_site_set_documents( site, name, docs );
		
		list = list->next;
	}
	
	ret = TRUE;

	priv->output = g_string_new( "<?xml version=\"1.0\" ?>\n<screemsession>" );

	list = screem_application_get_loaded_sites( priv->app );
	if( list ) {
		g_list_foreach( list,
				(GFunc)screem_session_save_site,
				session );
	} else {
		g_warning( "Default site missing!\n" );
	}
	
	list = screem_application_get_window_list( priv->app );
	if( list ) {
		g_list_foreach( list,
				(GFunc)screem_session_save_windows,
				session );
	} else {
		g_warning( "No windows!\n" );
	}

	screem_session_save_hint( session );
	screem_session_save_dialogs( session );
	
	g_string_append( priv->output, "</screemsession>" );
	
	file = get_session_filename( filename );
	ret = save_file( file, priv->output->str, 
			 GNOME_VFS_PERM_USER_READ |
			 GNOME_VFS_PERM_USER_WRITE,
			 FALSE, FALSE,
			 &error );
	if( error ) {
		g_printerr( "Failed to save session: %s\n",
				error->message );
		g_error_free( error );
	}
	g_free( file );
	
	g_string_free( priv->output, TRUE );
	priv->output = NULL;
	
	return ret;
}

void screem_session_set_dialog( ScreemSession *session, 
		const gchar *name, gint x, gint y, gint w, gint h )
{
	ScreemSessionPrivate *priv;
	GdkRectangle *rect;

	priv = session->priv;

	rect = g_hash_table_lookup( priv->dialogs, name );
	if( ! rect ) {
		rect = g_new0( GdkRectangle, 1 );
		g_hash_table_insert( priv->dialogs, g_strdup( name ), 
				rect );
	}
	rect->x = x;
	rect->y = y;
	if( w != -1 ) {
		rect->width = w;
	}
	if( h != -1 ) {
		rect->height = h;
	}
}

gboolean screem_session_get_dialog( ScreemSession *session, 
		const gchar *name, gint *x, gint *y, gint *w, gint *h )
{
	ScreemSessionPrivate *priv;
	GdkRectangle *rect;

	priv = session->priv;
	
	rect = g_hash_table_lookup( priv->dialogs, name );
	if( rect ) {
		if( x ) {
			*x = rect->x;
		}
		if( y ) {
			*y = rect->y;
		}
		if( w ) {
			*w = rect->width;
		}
		if( h ) { 
			*h = rect->height;
		}
	}

	return ( rect != NULL );
}

void screem_session_restore_dialog( ScreemSession *session,
		GtkWidget *widget )
{
	const gchar *role;
	gint x;
	gint y;
	gint w;
	gint h;
	
	g_return_if_fail( GTK_IS_WINDOW( widget ) );
	
	role = gtk_window_get_role( GTK_WINDOW( widget ) );

	g_return_if_fail( role != NULL );
	
	if( session && screem_session_get_dialog( session, role,
				&x, &y, &w, &h ) ) {
		gtk_window_move( GTK_WINDOW( widget ), x, y );
		gtk_window_resize( GTK_WINDOW( widget ), w, h );
	}
}

void screem_session_store_dialog( ScreemSession *session,
		GtkWidget *widget )
{
	const gchar *role;
	gint x;
	gint y;
	gint w;
	gint h;
	
	g_return_if_fail( GTK_IS_WINDOW( widget ) );
	
	role = gtk_window_get_role( GTK_WINDOW( widget ) );

	g_return_if_fail( role != NULL );
	
	if( session ) {
		gtk_window_get_size( GTK_WINDOW( widget ), &w, &h );
		gtk_window_get_position( GTK_WINDOW( widget ), &x, &y );
		screem_session_set_dialog( session, role, x, y, w, h );
	}
}

void screem_session_site_set_expanded( ScreemSession *session,
		const gchar *name, const gchar *uri,
		gboolean status )
{
	ScreemSessionPrivate *priv;
	GHashTable *expansion;
	
	g_return_if_fail( SCREEM_IS_SESSION( session ) );
	g_return_if_fail( name != NULL );
	g_return_if_fail( uri != NULL );
	
	priv = session->priv;

	expansion = g_hash_table_lookup( priv->site_expansion, name );
	if( ! expansion ) {
		expansion = g_hash_table_new_full( g_str_hash, 
				g_str_equal,
				(GDestroyNotify)g_free, NULL );
		g_hash_table_insert( priv->site_expansion,
				g_strdup( name ), expansion );
	}

	g_return_if_fail( expansion != NULL );

	if( g_hash_table_lookup( expansion, uri ) ) {
		g_return_if_fail( status != TRUE );
	
		g_hash_table_remove( expansion, uri );
		
	} else {
		g_return_if_fail( status != FALSE );

		g_hash_table_insert( expansion, g_strdup( uri ), 
				GINT_TO_POINTER( 1 ) );
	}
}

gboolean screem_session_site_get_expanded( ScreemSession *session,
		const gchar *name, const gchar *uri )
{
	ScreemSessionPrivate *priv;
	gboolean ret;
	GHashTable *expansion;
	
	g_return_val_if_fail( SCREEM_IS_SESSION( session ), FALSE );
	g_return_val_if_fail( name != NULL, FALSE );
	g_return_val_if_fail( uri != NULL, FALSE );
	
	priv = session->priv;
	
	expansion = g_hash_table_lookup( priv->site_expansion, name );
	
	ret = FALSE;
	if( expansion ) {
		ret = GPOINTER_TO_INT( g_hash_table_lookup( expansion, uri ) );
	}
	
	return ret;
}

